/*
**	wmcdrom 0.6
**
**	wmcdrom 0.1: (c) 2002 Brian Schau <brisse@ping.dk>
**	next versions: (c) 2004 Matteo Lucarelli <matteo.lucarelli@taolab.it>
**	wmcdrom 0.5-0.6: (c) 2007 Mateus Caruccio <mateus@caruccio.com>
**
**	wmcdrom is a dock application which provides a way to easily
**	mount and unmount removable filesystems.
**
**	See the file LICENSE included in this package for license terms.
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2 of the License, or
**	(at your option) any later version.
**
**	This program is distributed in the hope that it will be useful,
**	but WITHOUT ANY WARRANTY; without even the implied warranty of
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**	GNU General Public License for more details.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; if not, write to:
**
**		Free Software Foundation, Inc.
**		59 Temple Place - Suite 330
**		Boston, MA 02111-1307
**		USA
*/
#include <sys/mount.h>
#include <fstab.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <mntent.h>
#include <libgen.h>
#include <ctype.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xos.h>
#include <X11/xpm.h>
#include "cdromm.xpm"
#include "cdromu.xpm"
#include "dvdm.xpm"
#include "dvdu.xpm"
#include "floppym.xpm"
#include "floppyu.xpm"
#include "usbm.xpm"
#include "usbu.xpm"
#include "wait.xpm"
#include "error.xpm"

#ifndef PATH_MAX
#define PATH_MAX 256
#endif

/* macros */
#define PROGRAMNAME     "wmcdrom"
#define WMCDROM_VERSION "0.5"
#define DEFAULT_DEVICE  "/dev/cdrom"

/* commands */
//TODO: discover XServer prefix on Makefile or search commands on $PATH
#define MOUNTCMD "/bin/mount"
#define UMOUNTCMD "/bin/umount"
#define EJECTCMD "/usr/bin/eject"
#define FILEMANCMD "/usr/X11R6/bin/xterm" //not every one has mc
#define XMESSAGECMD "/usr/X11R6/bin/xmessage -center -default okay" //TODO: use failure icon ?

/* command flags */
#define HAS_MOUNTCMD    0x01
#define HAS_UMOUNTCMD   0x02
#define HAS_EJECTCMD    0x04
#define HAS_FILEMANCMD  0x08
#define HAS_DEVICE      0x10

/* UI */
#define DOCKWINDOWSIZE 64
#define ICONWINDOWSIZE 48
#define DEFAULT_TYPEICON "cd"

/* helper */
#define SEXIT(x) { fprintf(stderr, "%s: " x, PROGRAMNAME); exit(EXIT_FAILURE); }

typedef enum { DeviceNotMounted, DeviceMounted, DeviceError, unknown} deviceStatus;

/* protos */
static int resolveDevice(void);
//static void alarmHandler(int);
static void childHandler(int);
static void exitHandler(int);
static void xpmToPixmap(char **xpm, Pixmap *icon);
static void initDock(char* iconame);
static void eventLoop(void);
static deviceStatus mountStatus(void);
static void updateMountStatus(void);
static void flashButtonUp(void);
static void flashButtonDown(void);
static void leftButtonAction(void);
static void rightButtonAction(void);
static int commandDevice(char *);
void  parsecommandline(char *line, char **argv);

/* globals */
static deviceStatus oldDeviceStatus=unknown;
static Display *display;
static Window dockWindow, dockIconWindow;
static long iconEventMask = 0, dockEventMask = 0;
static Pixmap mountPix, unmountPix , waitPix, errorPix;
static Atom atomDeleteWindow;
static char *device = NULL, *cmdEject = NULL, *cmdMount = NULL, *cmdUmount = NULL, *cmdFileman = NULL;
static short cmdFlags = 0;
static char realDevice[PATH_MAX]; //resolve device's real name if it's a symlink
static int cdMntPoint = 0;
static int deviceError = 0;

/*
 * Collect child process status
 */
void childHandler(int signo)
{
	sigset_t mask;
	int status = 0;

	sigemptyset(&mask);
	sigaddset(&mask, SIGALRM);

	sigprocmask(SIG_BLOCK, &mask, NULL);
	wait(&status);
	sigprocmask(SIG_UNBLOCK, &mask, NULL);
};

/*
 * Force exit
 */
void alarmHandler(int signo)
{
	exit(EXIT_FAILURE);
};

/*
 * Clean up
 */
void exitHandler(int signo)
{
	if (cmdFlags & HAS_MOUNTCMD)
		free(cmdMount);

	if (cmdFlags & HAS_UMOUNTCMD)
		free(cmdUmount);
	
	if (cmdFlags & HAS_EJECTCMD)
		free(cmdEject);
	
	if (cmdFlags & HAS_FILEMANCMD)
		free(cmdFileman);

	if (cmdFlags & HAS_DEVICE)
		free(device);

	XKillClient(display, (XID)dockWindow);

	//some times (almost all) XKillClient is not niticed by eventLoop,
	//so we force a error exit.
	alarm(3);
};

/*
 * Install signal callbacks
 */
void InstallSigHandler()
{
	struct sigaction act;

	memset(&act, 0, sizeof(struct sigaction));
	act.sa_handler = alarmHandler;
	sigemptyset(&act.sa_mask);
	sigaction(SIGALRM, &act, NULL);

	memset(&act, 0, sizeof(struct sigaction));
	act.sa_handler = childHandler;
	sigemptyset(&act.sa_mask);

	sigaction(SIGCHLD, &act, NULL);

	memset(&act, 0, sizeof(struct sigaction));
	act.sa_handler = exitHandler;
	sigemptyset(&act.sa_mask);

	sigaction(SIGINT, &act, NULL);
	sigaction(SIGHUP, &act, NULL);
	sigaction(SIGTERM, &act, NULL);
};

/*
 * Show usage message
 */
void usage(void)
{
	//fprintf(stderr, "%s %s\n", PROGRAMNAME, WMCDROM_VERSION);
	fprintf(stderr, "Usage: %s [OPTIONS...] [DEVICE]\n\n", PROGRAMNAME);

	fprintf(stderr, "Mount, umount and eject DEVICE (default: %s).\n\n", DEFAULT_DEVICE);

	fprintf(stderr, "Left click first mounts (-m) then launch filemanager (-f, -c).\n");
	fprintf(stderr, "Right click first unmounts (-u) then eject (-e).\n");
	fprintf(stderr, "Middle click unmount and eject.\n\n");

	fprintf(stderr, "  -c           Change dir to mount point before execute command (see -f)\n");
	fprintf(stderr, "  -e PATH      Path to eject location (default: %s)\n", EJECTCMD);
	fprintf(stderr, "  -f PATH      Path to filemanager (default: %s)\n", FILEMANCMD);
	fprintf(stderr, "  -h           This help message\n");
	fprintf(stderr, "  -i TYPE      Icon type: cd, dvd, floppy, usb (default: %s)\n", DEFAULT_TYPEICON);
	fprintf(stderr, "  -m PATH      Path to mount location (default: %s)\n", MOUNTCMD);
	fprintf(stderr, "  -u PATH      Path to umount location (default: %s)\n", UMOUNTCMD);
	fprintf(stderr, "  -v           Show version and exit\n");
};

/*
**	main
*/
int main(int argc,char *argv[])
{
	int opt;
	char* typeIco = DEFAULT_TYPEICON;
	char has_typeIco = 0;

	while ((opt=getopt(argc, argv, "ce:m:u:i:f:vh"))!=-1) {
		switch (opt) {
		case 'v':
			printf("%s %s\n", PROGRAMNAME, WMCDROM_VERSION);
			exit(EXIT_SUCCESS);
		case 'c':
			cdMntPoint = 1;
			break;
		case 'e':
			cmdEject = strdup(optarg);
			cmdFlags |= HAS_EJECTCMD;
			break;
		case 'm':
			cmdMount = strdup(optarg);
			cmdFlags |= HAS_MOUNTCMD;
			break;
		case 'u':
			cmdUmount = strdup(optarg);
			cmdFlags |= HAS_UMOUNTCMD;
			break;
		case 'f':
			cmdFileman = strdup(optarg);
			cmdFlags |= HAS_FILEMANCMD;
			break;
		case 'i':
			typeIco = strdup(optarg);
			has_typeIco = 1;
			break;
		case 'h':
			usage();
			exit(EXIT_SUCCESS);
		default:
			usage();
			exit(EXIT_FAILURE);
		}
	}

	if (cmdEject==NULL) cmdEject=EJECTCMD;

	if (cmdMount==NULL) cmdMount=MOUNTCMD;

	if (cmdUmount==NULL) cmdUmount=UMOUNTCMD;

	if (cmdFileman==NULL) cmdFileman=FILEMANCMD;

	//TODO: automatic find [cdrom] device
	if (argc-optind==1) {
		device=strdup(argv[optind]);
		cmdFlags |= HAS_DEVICE;
	} else
		device = DEFAULT_DEVICE;

	if (!resolveDevice()) {
		usage();
		exit(EXIT_FAILURE);
	}

	fprintf(stderr, "%s: using device %s (%s)\n", PROGRAMNAME, device, realDevice);

	initDock(typeIco);

	if (!strcmp(typeIco,"cd")){
		xpmToPixmap(cdromm_xpm, &mountPix);
		xpmToPixmap(cdromu_xpm, &unmountPix);
	}else if (!strcmp(typeIco,"dvd")){
		xpmToPixmap(dvdm_xpm, &mountPix);
		xpmToPixmap(dvdu_xpm, &unmountPix);
	}else if (!strcmp(typeIco,"floppy")){
		xpmToPixmap(floppym_xpm, &mountPix);
		xpmToPixmap(floppyu_xpm, &unmountPix);
	}else if (!strcmp(typeIco,"usb")){
		xpmToPixmap(usbm_xpm, &mountPix);
		xpmToPixmap(usbu_xpm, &unmountPix);
	}

	if (has_typeIco)
		free(typeIco);

	xpmToPixmap(wait_xpm, &waitPix);
	xpmToPixmap(error_xpm, &errorPix);

	InstallSigHandler();

	updateMountStatus();
	XMapRaised(display, dockWindow);
	XFlush(display);

	eventLoop();

	XCloseDisplay(display);
	exit(EXIT_SUCCESS);
}

// parse a sentence in single words
void  parsecommandline(char *line, char **argv)
{
	while (*line != '\0') {
		while (isspace(*line)) *line++ = '\0';
		*argv++ = line;
		while (*line && !isspace(*line)) line++;
	}
	*argv = '\0';
}

/*
**	xpmToPixmap
**
**	Convert XPM to Pixmap
**
**	 -> xpm		Source xpm.
**	 -> icon	Where to store icon.
**
** TODO: Apply transparency mask
*/
static void
xpmToPixmap(char **xpm, Pixmap *icon)
{
	/*
	XpmAttributes attr;
	Pixmap mask;
	GC gc;

	attr.valuemask = XpmExactColors | XpmCloseness | XpmReturnPixels | XpmReturnExtensions;
	int error = XpmCreatePixmapFromData(display, dockWindow, xpm, icon, &mask, &attr);

	if (error == XpmSuccess) {
		XGCValues gcv;
		gcv.clip_mask = mask;
		gc = XCreateGC(display, *icon, GCClipMask, &gcv);
	}
	*/

	XpmCreatePixmapFromData(display, dockWindow, xpm, icon, NULL, NULL);
}

/*
**	initDock
**
**	Create dock dockWindow.
*/
static void initDock(char* iconame)
{
	XClassHint *classHint;
	XWMHints *hints;
	int black;
	char tmp[256];
	char *displayStr = NULL;

	if ((displayStr = getenv("DISPLAY")) == NULL)
		displayStr = "";

	if ((display=XOpenDisplay(displayStr)) == NULL)
		SEXIT("Cannot open display\n")

	atomDeleteWindow = XInternAtom(display, "WM_DELETE_WINDOW", False);
	black=BlackPixel(display, DefaultScreen(display));

	dockWindow=XCreateSimpleWindow(display, DefaultRootWindow(display), 0, 0, DOCKWINDOWSIZE, DOCKWINDOWSIZE, 0, black, black);

	if ((classHint=XAllocClassHint())==NULL)
		SEXIT("Failed to allocate class hint\n");

	strncpy(tmp,PROGRAMNAME,256);
	strncat(tmp,iconame,256);

	classHint->res_class=tmp;
	classHint->res_name=tmp;
	//classHint->res_class=PROGRAMNAME;
	//classHint->res_name=PROGRAMNAME;
	XSetClassHint(display, dockWindow, classHint);
	XFree(classHint);

	dockIconWindow=XCreateSimpleWindow(display, dockWindow, (DOCKWINDOWSIZE-ICONWINDOWSIZE)/2, (DOCKWINDOWSIZE-ICONWINDOWSIZE)/2, ICONWINDOWSIZE, ICONWINDOWSIZE, 0, black, black);

	XSetWMProtocols(display, dockWindow, &atomDeleteWindow, 1);

	if ((hints=XAllocWMHints())==NULL)
		SEXIT("Failed to allocate hints\n");

	hints->flags=StateHint|WindowGroupHint|IconWindowHint;
	hints->initial_state=WithdrawnState;
	hints->window_group=dockWindow;
	hints->icon_window=dockIconWindow;
	hints->icon_x=0;
	hints->icon_y=0;

	XSetWMHints(display, dockWindow, hints);
	XFree(hints);

	dockEventMask = ExposureMask | SubstructureRedirectMask | ButtonPressMask | ButtonReleaseMask|SubstructureNotifyMask|StructureNotifyMask;
	iconEventMask = ExposureMask | SubstructureRedirectMask | ButtonPressMask | ButtonReleaseMask|SubstructureNotifyMask|StructureNotifyMask;

	XSelectInput(display, dockWindow, dockEventMask);
	XSelectInput(display, dockIconWindow, iconEventMask);
	XSetCloseDownMode(display, DestroyAll);
}

/*
**	eventLoop
**
**	Handle events from X server.
*/
static void eventLoop(void)
{
	XEvent ev;
	time_t lastUpdate = 0, now = 0;
	const time_t updateTimeout = 1;
	int i = 0;

	for (;;) {
		if (XCheckMaskEvent(display, iconEventMask, &ev)) {

			switch (ev.type) {
				case Expose:
					resolveDevice();
					updateMountStatus();
					break;
				case ButtonPress:
					if (ev.xbutton.x<0 || ev.xbutton.x>ICONWINDOWSIZE || ev.xbutton.y<0 || ev.xbutton.y>ICONWINDOWSIZE)
						break;
					flashButtonUp();
					break;
				case ButtonRelease:
					if (ev.xbutton.x<0 || ev.xbutton.x>ICONWINDOWSIZE || ev.xbutton.y<0 || ev.xbutton.y>ICONWINDOWSIZE)
						break;
					if (ev.xbutton.button==1){
						flashButtonDown();
						leftButtonAction();
					} else if (ev.xbutton.button==3){
						flashButtonDown();
						rightButtonAction();
					} else if (ev.xbutton.button == 2) {
						flashButtonDown();
						commandDevice(cmdEject);
					}
					break;
				case ClientMessage:
					if (ev.xclient.data.l[0] == atomDeleteWindow) {
						XDestroyWindow(display, ev.xclient.window);

						exitHandler(0);
						return;
					}
				default:
					break;
					//return;
			}
		} else
			usleep(50000);

		now = time(NULL);

		if (now >= (lastUpdate + updateTimeout)) {
			lastUpdate = now;
			resolveDevice();
			updateMountStatus();
			deviceError = 0;
			i++;
		}
	}
}

/*
 * Recursively search for device real name
 */
static int resolveDevice(void)
{
	struct stat st;

	strncpy(realDevice, device, sizeof(realDevice));

	do {
		if (lstat(realDevice, &st) == -1)
			return 0;

		if (S_ISLNK(st.st_mode)) {
			char linkName[PATH_MAX];
			ssize_t linkLen;

			if ((linkLen = readlink(realDevice, linkName, sizeof(linkName))) > 0) {
				linkName[linkLen] = 0;

				if (linkName[0] != '/') {
					char *path = NULL;
					char *dir = NULL;

					if ((path = strdup(realDevice)) == NULL) {
						perror("strdup");
						exit(EXIT_FAILURE);
					}

					if ((dir = dirname(path)) == NULL) {
						perror("dirname");
						free(path);
						return 0;
					}

					strncpy(realDevice, dir, strlen(dir) + 1);
					strcat(realDevice, "/");
					strcat(realDevice, linkName);
					free(path);

				} else
					strncpy(realDevice, linkName, sizeof(realDevice));

			} else {
				perror("readlink");
				return 0;
			}

		} else if (S_ISBLK(st.st_mode)) {
			break;
		}

	} while (1);

	return 1;
}

/*
**	mountStatus
**
**	Return mount status for this device.
*/
static deviceStatus mountStatus(void)
{
	/* use a more portable way */
	FILE *fstab = NULL;
	struct mntent *ent = NULL;
	deviceStatus d = DeviceNotMounted;

	if (deviceError == 1)
		return DeviceError;

	if ((fstab = setmntent("/etc/mtab", "r")) != NULL) {
		while ((ent = getmntent(fstab)) != NULL) {
			if ((strcmp(ent->mnt_fsname, realDevice) == 0) || (strcmp(ent->mnt_fsname, realDevice) == 0)) {
				d = DeviceMounted;
				break;
			}
		}
		endmntent(fstab);
	} else
		perror("setmntent");

	return d;
}

/*
**	updateMountStatus
**
**	Update graphics in dockWindow.
*/
static void updateMountStatus(void)
{
	deviceStatus d=mountStatus();

	if (d==oldDeviceStatus)
		return;

	oldDeviceStatus=d;

	if (display) {
		switch (d) {
			case DeviceMounted:
				XSetWindowBackgroundPixmap(display, dockIconWindow, mountPix);
			break;
			case DeviceNotMounted:
				XSetWindowBackgroundPixmap(display, dockIconWindow, unmountPix);
			break;
			case DeviceError:
				XSetWindowBackgroundPixmap(display, dockIconWindow, errorPix);
			break;
			default: break;
		}
		XClearWindow(display, dockIconWindow);
		XFlush(display);
	}
}

/*
**	flashButton
**	Flash icon for click feedback
*/
static void flashButtonUp(void)
{
	if (display) {
		XSetWindowBackgroundPixmap(display, dockIconWindow, waitPix);
		XClearWindow(display, dockIconWindow);
		XFlush(display);
	}
}

static void flashButtonDown(void)
{
	deviceStatus d=mountStatus();

	if (display) {
		switch (d) {
			case DeviceMounted:
				XSetWindowBackgroundPixmap(display, dockIconWindow, mountPix);
			break;
			case DeviceNotMounted:
				XSetWindowBackgroundPixmap(display, dockIconWindow, unmountPix);
			break;
			case DeviceError:
				XSetWindowBackgroundPixmap(display, dockIconWindow, errorPix);
			break;
			default: break;
		}
		XClearWindow(display, dockIconWindow);
		XFlush(display);
	}
}

/*
**	leftButtonAction
**	Dispatch for left mousebutton click on icon.
*/
static void leftButtonAction(void)
{
	// mount or lounch fileman
	if (oldDeviceStatus==DeviceNotMounted)
		commandDevice(cmdMount);
	else if (oldDeviceStatus == DeviceMounted && fork() == 0) {
		char* argv[128];
		FILE *fstab = NULL;
		struct mntent *ent = NULL;

		parsecommandline(cmdFileman,argv);

		if (cdMntPoint) {
			if ((fstab = setmntent("/etc/fstab", "r")) != NULL) {
				while ((ent = getmntent(fstab)) != NULL) {
					if (strcmp(ent->mnt_fsname, device) == 0) {
						if (chdir(ent->mnt_dir) == -1)
							perror("chdir");
						break;
					}
				}
				endmntent(fstab);
			} else
				perror("setmntent");
		}

		if (execvp(*argv, argv) < 0) {
			fprintf(stderr, "%s: launching fileman failed!\n", PROGRAMNAME);
			exit(EXIT_FAILURE);
		}
	}
}

/*
**	rightButtonAction
**
**	Dispatch for right mousebutton click on icon.
*/
static void rightButtonAction(void)
{
	// umount or eject
	if (oldDeviceStatus==DeviceNotMounted || oldDeviceStatus == DeviceError)
		commandDevice(cmdEject);
	else if (oldDeviceStatus == DeviceMounted)
		commandDevice(cmdUmount);
}

/*
**  execute command on Device
**	 -> com		Command.
*/
static int commandDevice(char *com)
{
	int ret=-1;
	char *str = (char *)calloc(strlen(com) + strlen(device) + 7, 1);
	FILE *filep = NULL;
	char com_out[getpagesize()];
	const int len = getpagesize();

	if (display) {
		XSetWindowBackgroundPixmap(display, dockIconWindow, waitPix);
		XClearWindow(display, dockIconWindow);
		XFlush(display);
	}

	if (str) {
		sprintf(str,"%s %s 2>&1",com,device);
		filep = popen(str, "r");

		if (filep != NULL) {
			setbuf(filep, NULL);

			if (fgets(com_out, len, filep) != NULL) {
				fprintf(stderr, "%s: %s error: %s", PROGRAMNAME, com, com_out);
				deviceError = 1;
			} else
				deviceError = 0;

			pclose(filep);
		}

		free(str);
	}

	flashButtonDown();

	return ret;
}

